<?php

namespace Drupal\search_api_exactmatch_boost\Plugin\search_api\processor;

use Drupal;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\Item\Item;
use Drupal\search_api\Plugin\PluginFormTrait;
use Drupal\search_api\Processor\FieldsProcessorPluginBase;
use Drupal\search_api\Query\ResultSetInterface;
use Drupal\search_api\SearchApiException;

/**
 * Puts elements that exactly match to the top of the results list
 *
 * This processor puts elements that exactly match the text in a string index
 * field to the top of the results list. If a pager is used there maybe more
 * items on the first page and less items on following pages. Exact matches are
 * added to the first page, and removed from the regular result set afterwards.
 * If they appear in following pages they will be removed there.
 *
 * @SearchApiProcessor(
 *   id = "exactmatchboost",
 *   label = @Translation("Exact match boosting"),
 *   description = @Translation("Puts elements that exactly match to the top of the results list. Will only run on string index fields. Pager count may differ."),
 *   stages = {
 *     "postprocess_query" = 0,
 *   }
 * )
 */
class ExactMatchBoost extends FieldsProcessorPluginBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * Search server instance
   *
   * @var \Drupal\search_api\ServerInterface|null
   */
  protected $searchServer;

  /**
   * Transliteration processor if activated
   *
   * @var \Drupal\search_api\Plugin\search_api\processor\Transliteration
   */
  protected $transliterationProcessor;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    try {
      $this->searchServer = $this->getIndex()->getServerInstance();
    } catch (SearchApiException $e) {
      $this->searchServer = FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    if ($this->searchServer) {
      $form['exakt_match_boost'] = [
        'hint' => [
          '#type' => 'markup',
          '#markup' => $this->t('will only work on string fields')
        ]
      ];
      $form = parent::buildConfigurationForm($form, $form_state);
      if ($this->searchServer->getBackendId() != 'search_api_db') {
        return [
          '#type' => 'markup',
          '#markup' => $this->t('This processor does only work with Search API DB')
        ];
      }
      $form['all_fields']['#access'] = FALSE;
      $fields = $this->getIndex()->getFields();
      foreach ($fields as $field) {
        if ($field->getType() !== 'string') {
          unset($form['fields']['#options'][$field->getFieldIdentifier()]);
        }
      }
    }
    else {
      $form['exakt_match_boost'] = [
        'hint' => [
          '#type' => 'markup',
          '#markup' => $this->t('You need to add this index to a server first')
        ]
      ];
    }
    return $form;
  }

  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    if ($this->searchServer->getBackendId() != 'search_api_db') {
      $form_state->setError($form, $this->t('This processor does only work with Search API DB'));
    }
    parent::validateConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $configuration = $this->getConfiguration();
    $value_fields = $values['fields'] ?? [];
    $configuration_fields = $configuration['fields'] ?? [];
    if(!empty(array_diff($value_fields, $configuration_fields))) {
      $form_state->set('processors_changed', TRUE);
    }
    $this->setConfiguration($values);
  }

  /**
   * {@inheritdoc}
   */
  public function postprocessSearchResults(ResultSetInterface $results) {
    $changeresult = FALSE;

    if ($this->searchServer->getBackendId() == 'search_api_db') {
      $query = $results->getQuery();
      $searchkeys = $query->getOriginalKeys();
      if (!empty($searchkeys)) {
        /* @var \Drupal\views\ViewExecutable $view */
        $view = $query->getOption('search_api_view');
        /* @var \Drupal\search_api_db\Plugin\search_api\backend\Database $backend */
        $backend = $this->searchServer->getBackend();
        $database = $backend->getDatabase();
        $items = [];
        try {
          $this->transliterationProcessor = $this->index->getProcessor('transliteration');
        } catch (SearchApiException $e) {
          $this->transliterationProcessor = FALSE;
        }
        foreach ($this->configuration['fields'] as $key => $field) {
          if ($field !== 0) {
            if ($this->transliterationProcessor) {
              $translitConfig = $this->transliterationProcessor->getConfiguration();

              if (in_array($field, $translitConfig['fields'])) {
                $searchkeys = $this->transliterationProcessor->getTransliterator()
                  ->transliterate($searchkeys, $this->transliterationProcessor->getLangcode());
              }
            }
            $dbtable = $this->searchServer->getBackendId() . '_' . $this->index->id() . '_' . $field;
            $dbquery = $database->select($dbtable, 'dt');
            $dbquery->condition('dt.value', $searchkeys);
            $dbquery->fields('dt', ['item_id']);
            $dbresults = $dbquery->execute();
            foreach ($dbresults as $dbresult) {
              $items[$dbresult->item_id] = new Item($this->index, $dbresult->item_id);
            }
          }
        }
        if (!empty($items)) {
          $changeresult = TRUE;
          $oldresults = $results->getResultItems();
          if (empty($view) || $view->usePager() === FALSE || $view->getCurrentPage() === 0) {
            $newresults = array_merge($items, $oldresults);
          }
          else {
            foreach ($items as $key => $item) {
              unset($oldresults[$key]);
            }
            $newresults = $oldresults;
          }
        }
      }
    }
    if ($changeresult) {
      $results->setResultItems($newresults);
    }
    parent::postprocessSearchResults($results);
  }

}
